import BusX
import Combine

extension MeetupId {
  final class Debounce<Src, Dst> {
    let v = PassthroughSubject<Src, Never>()
    var subscriptions = [AnyCancellable]()

    init(
      _ sec: Double,
      _ handler: @escaping ((Src) -> Dst?),
      _ src: String,
      _ dst: String
    ) {
      Bus.receive(
        [src],
        { [weak self] _, v in self?.v.send(v) },
        sub: &subscriptions
      )

      Bus.send(
        dst,
        v
          .debounce(for: .seconds(sec), scheduler: DispatchQueue.main)
          .compactMap { (v: Src) in handler(v) }
          .eraseToAnyPublisher(),
        sub: &subscriptions
      )
    }
  }
}
